from __future__ import annotations

from murfey.util.file_monitor import Monitor


def test_empty_directory_does_nothing(tmp_path):
    monitor = Monitor(tmp_path)
    monitor.process(in_thread=True, sleep=0.1)
    monitor.stop()
    monitor.wait()
    assert not monitor._timed_cache
    assert not monitor._out.get()
    assert monitor._out.empty()


def test_directory_with_a_file_finds_that_file(tmp_path):
    (tmp_path / "empty_file.txt").touch()
    monitor = Monitor(tmp_path)
    monitor.process(in_thread=True, sleep=0.1)
    monitor.stop()
    monitor.wait()
    assert len(monitor._timed_cache.keys()) == 1
    assert monitor._out.get() == [tmp_path / "empty_file.txt"]
    assert not monitor._out.get()
    assert monitor._out.empty()


def test_directory_with_an_added_file(tmp_path):
    (tmp_path / "empty_file.txt").touch()
    monitor = Monitor(tmp_path)
    monitor.process(in_thread=True, sleep=0.1)
    (tmp_path / "another_empty_file.txt").touch()
    monitor.stop()
    monitor.wait()
    assert len(monitor._timed_cache.keys()) == 2



import os
from unittest.mock import patch

import yaml
from cryptography.fernet import Fernet

from murfey.cli.decrypt_db_password import run
from murfey.server.config import get_machine_config


def test_decrypt_password(capsys, tmp_path):
    machine_config = get_machine_config()
    crypto_key = Fernet.generate_key()
    machine_config.crypto_key = crypto_key.decode()
    with open(tmp_path / "config.yaml", "w") as cfg:
        yaml.dump({"m01": machine_config.dict()}, cfg)
    os.environ["MURFEY_MACHINE_CONFIGURATION"] = str(tmp_path / "config.yaml")
    os.environ["BEAMLINE"] = "m01"
    password = "abcd"
    f = Fernet(crypto_key)
    encrypted_password = f.encrypt(password.encode("ascii")).decode()
    with patch("argparse._sys.argv", ["murfey.decrypt_password", encrypted_password]):
        run()
    captured = capsys.readouterr()
    assert captured.out.replace("\n", "").replace("\r\n", "") == password



"""
A progressbar widget that uses block characters rather than a thin
horizontal line. This makes it more visible on Windows.
"""

from __future__ import annotations

import math
from functools import lru_cache
from typing import List

import rich.progress
import rich.progress_bar
from rich.color import Color, blend_rgb
from rich.color_triplet import ColorTriplet
from rich.console import Console, ConsoleOptions, RenderResult
from rich.segment import Segment
from rich.style import Style

PULSE_SIZE = rich.progress_bar.PULSE_SIZE


class BlockProgressBar(rich.progress_bar.ProgressBar):
    @lru_cache(maxsize=16)
    def _get_pulse_segments(
        self,
        fore_style: Style,
        back_style: Style,
        color_system: str,
        no_color: bool,
        ascii: bool = False,
    ) -> List[Segment]:
        """Get a list of segments to render a pulse animation.

        Returns:
            List[Segment]: A list of segments, one segment per character.
        """
        bar = "-" if ascii else "█"
        segments: List[Segment] = []
        if color_system not in ("standard", "eight_bit", "truecolor") or no_color:
            segments += [Segment(bar, fore_style)] * (PULSE_SIZE // 2)
            segments += [Segment(" " if no_color else bar, back_style)] * (
                PULSE_SIZE - (PULSE_SIZE // 2)
            )
            return segments

        append = segments.append
        fore_color = (
            fore_style.color.get_truecolor()
            if fore_style.color
            else ColorTriplet(255, 0, 255)
        )
        back_color = (
            back_style.color.get_truecolor()
            if back_style.color
            else ColorTriplet(0, 0, 0)
        )
        cos = math.cos
        pi = math.pi
        _Segment = Segment
        _Style = Style
        from_triplet = Color.from_triplet

        for index in range(PULSE_SIZE):
            position = index / PULSE_SIZE
            fade = 0.5 + cos(position * pi * 2) / 2.0
            color = blend_rgb(fore_color, back_color, cross_fade=fade)
            append(_Segment(bar, _Style(color=from_triplet(color))))
        return segments

    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:

        width = min(self.width or options.max_width, options.max_width)
        ascii = options.ascii_only
        if self.pulse:
            yield from self._render_pulse(console, width, ascii=ascii)
            return

        completed = min(self.total, max(0, self.completed))

        bar = "-" if ascii else "█"
        half_bar_right = " " if ascii else "▌"
        half_bar_left = " " if ascii else "▐"
        complete_halves = (
            int(width * 2 * completed / self.total) if self.total else width * 2
        )
        bar_count = complete_halves // 2
        half_bar_count = complete_halves % 2
        style = console.get_style(self.style)
        complete_style = console.get_style(
            self.complete_style if self.completed < self.total else self.finished_style
        )
        _Segment = Segment
        if bar_count:
            yield _Segment(bar * bar_count, complete_style)
        if half_bar_count:
            yield _Segment(half_bar_right * half_bar_count, complete_style)

        if not console.no_color:
            remaining_bars = width - bar_count - half_bar_count
            if remaining_bars and console.color_system is not None:
                if not half_bar_count and bar_count:
                    yield _Segment(half_bar_left, style)
                    remaining_bars -= 1
                if remaining_bars:
                    yield _Segment(bar * remaining_bars, style)


class BlockBarColumn(rich.progress.BarColumn):
    def render(self, task: rich.progress.Task) -> rich.progress_bar.ProgressBar:
        """Gets a progress bar widget for a task."""
        return BlockProgressBar(
            total=max(0, task.total),
            completed=max(0, task.completed),
            width=None if self.bar_width is None else max(1, self.bar_width),
            pulse=not task.started,
            animation_time=task.get_time(),
            style=self.style,
            complete_style=self.complete_style,
            finished_style=self.finished_style,
            pulse_style=self.pulse_style,
        )


if __name__ == "__main__":  # pragma: no cover
    from rich.theme import Theme

    custom_theme = Theme(
        {
            "bar.complete": "rgb(249,38,249)",
            "bar.finished": "rgb(31,156,31)",
            "bar.pulse": "rgb(249,38,249)",
        }
    )
    console = Console(theme=custom_theme)
    steps = 400
    arguments = {
        "width": 50,
    }
    bars = [
        BlockProgressBar(**arguments, total=steps),
        rich.progress_bar.ProgressBar(**arguments, total=steps),
        BlockProgressBar(**arguments, pulse=True),
        rich.progress_bar.ProgressBar(**arguments, pulse=True),
    ]

    import time

    console.show_cursor(False)
    for n in range(0, steps + 1, 1):
        for bar in bars:
            bar.update(n)
            console.print(bar)
            console.print(" ", end="")
        console.file.write("\r")
        time.sleep(3 / steps)
    console.show_cursor(True)
    console.print()



from __future__ import annotations

import logging
import queue
import threading
from pathlib import Path
from typing import Type

from murfey.client.context import Context
from murfey.client.contexts.spa import SPAContext, SPAModularContext
from murfey.client.contexts.spa_metadata import SPAMetadataContext
from murfey.client.contexts.tomo import TomographyContext
from murfey.client.instance_environment import MurfeyInstanceEnvironment
from murfey.client.rsync import RSyncerUpdate, TransferResult
from murfey.client.tui.forms import FormDependency
from murfey.util import Observer, get_machine_config
from murfey.util.models import PreprocessingParametersTomo, ProcessingParametersSPA

logger = logging.getLogger("murfey.client.analyser")

spa_form_dependencies: dict = {
    "use_cryolo": FormDependency(
        dependencies={"estimate_particle_diameter": False}, trigger_value=False
    ),
    "estimate_particle_diameter": FormDependency(
        dependencies={
            "use_cryolo": True,
            "boxsize": "None",
            "small_boxsize": "None",
            "mask_diameter": "None",
            "particle_diameter": "None",
        },
        trigger_value=True,
    ),
}


class Analyser(Observer):
    def __init__(
        self,
        basepath_local: Path,
        environment: MurfeyInstanceEnvironment | None = None,
        force_mdoc_metadata: bool = False,
        limited: bool = False,
    ):
        super().__init__()
        self._basepath = basepath_local.absolute()
        self._limited = limited
        self._experiment_type = ""
        self._acquisition_software = ""
        self._role = ""
        self._extension: str = ""
        self._unseen_xml: list = []
        self._context: Context | None = None
        self._batch_store: dict = {}
        self._environment = environment
        self._force_mdoc_metadata = force_mdoc_metadata
        self.parameters_model: (
            Type[ProcessingParametersSPA] | Type[PreprocessingParametersTomo] | None
        ) = None

        self.queue: queue.Queue = queue.Queue()
        self.thread = threading.Thread(name="Analyser", target=self._analyse)
        self._stopping = False
        self._halt_thread = False
        self._murfey_config = (
            get_machine_config(str(environment.url.geturl()), demo=environment.demo)
            if environment
            else {}
        )

    def _find_extension(self, file_path: Path):
        if (
            required_substrings := self._murfey_config.get(
                "data_required_substrings", {}
            )
            .get(self._acquisition_software, {})
            .get(file_path.suffix)
        ):
            if not any(r in file_path.name for r in required_substrings):
                return []

        if (
            file_path.suffix in (".mrc", ".tiff", ".tif", ".eer")
            and not self._extension
        ):
            logger.info(f"File extension determined: {file_path.suffix}")
            self._extension = file_path.suffix
        elif (
            file_path.suffix in (".tiff", ".tif", ".eer")
            and self._extension != file_path.suffix
        ):
            logger.info(f"File extension re-evaluated: {file_path.suffix}")
            self._extension = file_path.suffix

    def _find_context(self, file_path: Path) -> bool:
        split_file_name = file_path.name.split("_")
        if split_file_name:
            if split_file_name[0].startswith("FoilHole"):
                if not self._context:
                    logger.info("Acquisition software: EPU")
                    if self._environment:
                        try:
                            cfg = get_machine_config(
                                str(self._environment.url.geturl()),
                                demo=self._environment.demo,
                            )
                        except Exception as e:
                            logger.warning(f"exception encountered: {e}")
                            cfg = {}
                    else:
                        cfg = {}
                    self._context = (
                        SPAModularContext("epu", self._basepath)
                        if cfg.get("modular_spa")
                        else SPAContext("epu", self._basepath)
                    )
                self.parameters_model = ProcessingParametersSPA
                if not self._role:
                    self._role = "detector"
                return True
            if (
                split_file_name[0] == "Position"
                or "[" in file_path.name
                or "Fractions" in split_file_name[-1]
                or "fractions" in split_file_name[-1]
            ):
                if not self._context:
                    logger.info("Acquisition software: tomo")
                    self._context = TomographyContext("tomo", self._basepath)
                    self.parameters_model = PreprocessingParametersTomo
                if not self._role:
                    if (
                        "Fractions" in split_file_name[-1]
                        or "fractions" in split_file_name[-1]
                    ):
                        self._role = "detector"
                    elif (
                        file_path.suffix == ".mdoc"
                        or file_path.with_suffix(".mdoc").is_file()
                    ):
                        self._role = "microscope"
                    else:
                        self._role = "detector"
                return True
            if file_path.suffix in (".mrc", ".tiff", ".tif", ".eer"):
                if any(p in file_path.parts for p in ("Batch", "SearchMaps")):
                    return False
                if file_path.with_suffix(".jpg").is_file():
                    return False
                if (
                    len(
                        list(
                            file_path.parent.glob(
                                f"{file_path.name}*{file_path.suffix}"
                            )
                        )
                    )
                    > 1
                ):
                    # This covers the case of ignoring the averaged movies written out by the Falcon
                    return False
                self._context = TomographyContext("serialem", self._basepath)
                self.parameters_model = PreprocessingParametersTomo
                if not self._role:
                    if "Frames" in file_path.parts:
                        self._role = "detector"
                    else:
                        self._role = "microscope"
                return True
        return False

    def post_transfer(self, transferred_file: Path):
        try:
            if self._context:
                self._context.post_transfer(
                    transferred_file, role=self._role, environment=self._environment
                )
        except Exception as e:
            logger.error(f"An exception was encountered post transfer: {e}")

    def _analyse(self):
        logger.info("Analyser thread started")
        mdoc_for_reading = None
        while not self._halt_thread:
            transferred_file = self.queue.get()
            if not transferred_file:
                self._halt_thread = True
                continue
            if self._limited:
                if (
                    "Metadata" in transferred_file.parts
                    or transferred_file.name == "EpuSession.dm"
                    and not self._context
                ):
                    self._context = SPAMetadataContext("epu", self._basepath)
                self.post_transfer(transferred_file)
            else:
                dc_metadata = {}
                if (
                    self._force_mdoc_metadata
                    and transferred_file.suffix == ".mdoc"
                    or mdoc_for_reading
                ):
                    if self._context:
                        try:
                            dc_metadata = self._context.gather_metadata(
                                mdoc_for_reading or transferred_file,
                                environment=self._environment,
                            )
                        except KeyError as e:
                            logger.error(
                                f"Metadata gathering failed with a key error for key: {e.args[0]}"
                            )
                            raise e
                        if not dc_metadata:
                            mdoc_for_reading = None
                    elif transferred_file.suffix == ".mdoc":
                        mdoc_for_reading = transferred_file
                if not self._context:
                    self._find_extension(transferred_file)
                    found = self._find_context(transferred_file)
                    if not found:
                        # logger.warning(
                        #     f"Context not understood for {transferred_file}, stopping analysis"
                        # )
                        self.queue.task_done()
                        continue
                    elif self._extension:
                        logger.info(f"Context found successfully: {self._role}")
                        try:
                            self._context.post_first_transfer(
                                transferred_file,
                                role=self._role,
                                environment=self._environment,
                            )
                        except Exception as e:
                            logger.warning(f"exception encountered {e}")
                        if self._role == "detector":
                            if not dc_metadata:
                                try:
                                    dc_metadata = self._context.gather_metadata(
                                        (
                                            transferred_file.with_suffix(".mdoc")
                                            if self._context._acquisition_software
                                            == "serialem"
                                            else self._xml_file(transferred_file)
                                        ),
                                        environment=self._environment,
                                    )
                                except NotImplementedError:
                                    dc_metadata = {}
                                except KeyError as e:
                                    logger.error(
                                        f"Metadata gathering failed with a key error for key: {e.args[0]}"
                                    )
                                    raise e
                            if not dc_metadata or not self._force_mdoc_metadata:
                                self._unseen_xml.append(transferred_file)
                            else:
                                self._unseen_xml = []
                                if dc_metadata.get("file_extension"):
                                    self._extension = dc_metadata["file_extension"]
                                else:
                                    dc_metadata["file_extension"] = self._extension
                                dc_metadata["acquisition_software"] = (
                                    self._context._acquisition_software
                                )
                                self.notify(
                                    {
                                        "form": dc_metadata,
                                        "dependencies": (
                                            spa_form_dependencies
                                            if isinstance(self._context, SPAContext)
                                            or isinstance(
                                                self._context, SPAModularContext
                                            )
                                            else {}
                                        ),
                                    }
                                )
                elif not self._extension or self._unseen_xml:
                    self._find_extension(transferred_file)
                    if self._extension:
                        logger.info(
                            f"Context found successfully: {self._role}, {transferred_file}"
                        )
                        try:
                            self._context.post_first_transfer(
                                transferred_file,
                                role=self._role,
                                environment=self._environment,
                            )
                        except Exception as e:
                            logger.info(f"exception encountered {e}")
                        if self._role == "detector":
                            if not dc_metadata:
                                try:
                                    dc_metadata = self._context.gather_metadata(
                                        mdoc_for_reading
                                        or self._xml_file(transferred_file),
                                        environment=self._environment,
                                    )
                                except KeyError as e:
                                    logger.error(
                                        f"Metadata gathering failed with a key error for key: {e.args[0]}"
                                    )
                                    raise e
                            if not dc_metadata or not self._force_mdoc_metadata:
                                mdoc_for_reading = None
                                self._unseen_xml.append(transferred_file)
                            if dc_metadata:
                                self._unseen_xml = []
                                if dc_metadata.get("file_extension"):
                                    self._extension = dc_metadata["file_extension"]
                                else:
                                    dc_metadata["file_extension"] = self._extension
                                dc_metadata["acquisition_software"] = (
                                    self._context._acquisition_software
                                )
                                self.notify(
                                    {
                                        "form": dc_metadata,
                                        "dependencies": (
                                            spa_form_dependencies
                                            if isinstance(self._context, SPAContext)
                                            or isinstance(
                                                self._context, SPAModularContext
                                            )
                                            else {}
                                        ),
                                    }
                                )
                elif isinstance(
                    self._context,
                    (TomographyContext, SPAModularContext, SPAMetadataContext),
                ):
                    self.post_transfer(transferred_file)
            self.queue.task_done()

    def _xml_file(self, data_file: Path) -> Path:
        if not self._environment:
            return data_file.with_suffix(".xml")
        file_name = f"{'_'.join(p for p in data_file.stem.split('_')[:-1])}.xml"
        data_directories = self._murfey_config.get("data_directories", {})
        for dd in data_directories.keys():
            if str(data_file).startswith(dd):
                base_dir = Path(dd)
                mid_dir = data_file.relative_to(dd).parent
                break
        else:
            return data_file.with_suffix(".xml")
        return base_dir / self._environment.visit / mid_dir / file_name

    def enqueue(self, rsyncer: RSyncerUpdate):
        if not self._stopping and rsyncer.outcome == TransferResult.SUCCESS:
            absolute_path = (self._basepath / rsyncer.file_path).resolve()
            self.queue.put(absolute_path)

    def start(self):
        if self.thread.is_alive():
            raise RuntimeError("Analyser already running")
        if self._stopping:
            raise RuntimeError("Analyser has already stopped")
        logger.info(f"Analyser thread starting for {self}")
        self.thread.start()

    def stop(self):
        logger.debug("Analyser thread stop requested")
        self._stopping = True
        self._halt_thread = True
        try:
            if self.thread.is_alive():
                self.queue.put(None)
                self.thread.join()
        except Exception as e:
            logger.debug(f"Exception encountered while stopping analyser: {e}")
        logger.debug("Analyser thread stop completed")



import json

from murfey.util import set_default_acquisition_output


def test_set_default_acquisition_output_normal_operation(tmp_path):
    output_dir = tmp_path / "settings.json"
    settings_json = {
        "a": {
            "b": {"data_dir": str(tmp_path)},
            "c": {
                "d": 1,
            },
        }
    }
    with open(output_dir, "w") as sf:
        json.dump(settings_json, sf)
    set_default_acquisition_output(
        tmp_path / "visit", {str(tmp_path / "settings.json"): ["a", "b", "data_dir"]}
    )
    assert (tmp_path / "_murfey_settings.json").is_file()
    with open(output_dir, "r") as sf:
        data = json.load(sf)
    assert data["a"]["b"]["data_dir"] == str(tmp_path / "visit")
    assert data["a"]["c"]["d"] == 1
    with open(output_dir.parent / "_murfey_settings.json", "r") as sf:
        data = json.load(sf)
    assert data["a"]["b"]["data_dir"] == str(tmp_path)
    assert data["a"]["c"]["d"] == 1


def test_set_default_acquisition_output_no_file_copy(tmp_path):
    output_dir = tmp_path / "settings.json"
    settings_json = {
        "a": {
            "b": {"data_dir": str(tmp_path)},
            "c": {
                "d": 1,
            },
        }
    }
    with open(output_dir, "w") as sf:
        json.dump(settings_json, sf)
    set_default_acquisition_output(
        tmp_path / "visit",
        {str(tmp_path / "settings.json"): ["a", "b", "data_dir"]},
        safe=False,
    )
    assert not (tmp_path / "_murfey_settings.json").is_file()
    with open(output_dir, "r") as sf:
        data = json.load(sf)
    assert data["a"]["b"]["data_dir"] == str(tmp_path / "visit")
    assert data["a"]["c"]["d"] == 1



from __future__ import annotations

import argparse
import configparser

# import json
import logging
import os
import platform
import shutil
import sys
import time
import webbrowser
from datetime import datetime
from pathlib import Path
from queue import Queue
from typing import List, Literal
from urllib.parse import ParseResult, urlparse

import requests

# from multiprocessing import Process, Queue
from rich.prompt import Confirm

import murfey.client.rsync
import murfey.client.update
import murfey.client.watchdir
import murfey.client.websocket
from murfey.client.customlogging import CustomHandler, DirectableRichHandler
from murfey.client.instance_environment import MurfeyInstanceEnvironment
from murfey.client.tui.app import MurfeyTUI
from murfey.client.tui.status_bar import StatusBar
from murfey.util import _get_visit_list

# from asyncio import Queue


# from rich.prompt import Prompt


log = logging.getLogger("murfey.client")


def _enable_webbrowser_in_cygwin():
    """Helper function to make webbrowser.open() work in CygWin"""
    if "cygwin" in platform.system().lower() and shutil.which("cygstart"):
        webbrowser.register("cygstart", None, webbrowser.GenericBrowser("cygstart"))


def _check_for_updates(
    server: ParseResult, install_version: None | Literal[True] | str
):
    if install_version is True:
        # User requested installation of the newest version
        try:
            murfey.client.update.check(server, force=True)
            print("\nYou are already running the newest version of Murfey")
            exit()
        except Exception as e:
            exit(f"Murfey update check failed with {e}")

    if install_version:
        # User requested installation of a specific version
        if murfey.client.update.install_murfey(server, install_version):
            print(f"\nMurfey has been updated to version {install_version}")
            exit()
        else:
            exit("Error occurred while updating Murfey")

    # Otherwise run a routine update check to ensure client and server are compatible
    try:
        murfey.client.update.check(server)
    except Exception as e:
        print(f"Murfey update check failed with {e}")


def run():
    config = read_config()
    try:
        server_routing = config["ServerRouter"]
    except KeyError:
        server_routing = {}
    server_routing_prefix_found = False
    if server_routing:
        for path_prefix, server in server_routing.items():
            if str(Path.cwd()).startswith(path_prefix):
                known_server = server
                server_routing_prefix_found = True
                break
            else:
                known_server = None
    else:
        known_server = config["Murfey"].get("server")

    parser = argparse.ArgumentParser(description="Start the Murfey client")
    parser.add_argument(
        "--server",
        metavar="HOST:PORT",
        type=str,
        help=f"Murfey server to connect to ({known_server})",
        default=known_server,
    )
    parser.add_argument("--visit", help="Name of visit")
    parser.add_argument(
        "--source", help="Directory to transfer files from", type=Path, default="."
    )
    parser.add_argument(
        "--destination",
        help="Directory to transfer files to (syntax: 'data/2022/cm31093-2/tmp/murfey')",
    )
    parser.add_argument(
        "--update",
        metavar="VERSION",
        nargs="?",
        default=None,
        const=True,
        help="Update Murfey to the newest or to a specific version",
    )
    parser.add_argument(
        "--demo",
        action="store_true",
    )
    parser.add_argument(
        "--appearance-time",
        type=float,
        default=-1,
        help="Only consider top level directories that have appeared more recently than this many hours ago",
    )
    parser.add_argument(
        "--fake-dc",
        action="store_true",
        default=False,
        help="Do not perform data collection related calls to API (avoids database inserts)",
    )
    parser.add_argument(
        "--time-based-transfer",
        action="store_true",
        help="Transfer new files",
    )
    parser.add_argument(
        "--no-transfer",
        action="store_true",
        help="Avoid actually transferring files",
    )
    parser.add_argument(
        "--debug",
        action="store_true",
        help="Turn on debugging logs",
    )
    parser.add_argument(
        "--local",
        action="store_true",
        default=False,
        help="Perform rsync transfers locally rather than remotely",
    )
    parser.add_argument(
        "--ignore-mdoc-metadata",
        action="store_true",
        default=False,
        help="Do not attempt to read metadata from all mdoc files",
    )
    parser.add_argument(
        "--remove-files",
        action="store_true",
        default=False,
        help="Remove source files immediately after their transfer",
    )
    parser.add_argument(
        "--relax",
        action="store_true",
        default=False,
        help="Relax the condition that the source directory needs to be recognised from the configuration",
    )
    parser.add_argument(
        "--name",
        type=str,
        default="",
        help="Name of Murfey session to be created",
    )
    parser.add_argument(
        "--skip-existing-processing",
        action="store_true",
        default=False,
        help="Do not trigger processing for any data directories currently on disk (you may have started processing for them in a previous murfey run)",
    )

    args = parser.parse_args()

    if not args.server:
        exit("Murfey server not set. Please run with --server host:port")
    if not args.server.startswith(("http://", "https://")):
        if "://" in args.server:
            exit("Unknown server protocol. Only http:// and https:// are allowed")
        args.server = f"http://{args.server}"

    if args.remove_files:
        remove_prompt = Confirm.ask(
            f"Are you sure you want to remove files from {args.source or Path('.').resolve()}?"
        )
        if not remove_prompt:
            exit("Exiting")

    murfey_url = urlparse(args.server, allow_fragments=False)
    if args.server != known_server:
        # New server specified. Verify that it is real
        print(f"Attempting to connect to new server {args.server}")
        try:
            murfey.client.update.check(murfey_url, install=False)
        except Exception as e:
            exit(f"Could not reach Murfey server at {args.server!r} - {e}")

        # If server is reachable then update the configuration
        config["Murfey"]["server"] = args.server
        write_config(config)

    # If user requested installation of a specific or a newer version then
    # make that happen, otherwise ensure client and server are compatible and
    # update if necessary.
    _check_for_updates(server=murfey_url, install_version=args.update)

    if args.no_transfer:
        log.info("No files will be transferred as --no-transfer flag was specified")

    from pprint import pprint

    ongoing_visits = []
    if args.visit:
        ongoing_visits = [args.visit]
    elif server_routing_prefix_found:
        for part in Path.cwd().parts:
            if "-" in part:
                ongoing_visits = [part]
                break
    if not ongoing_visits:
        print("Ongoing visits:")
        ongoing_visits = _get_visit_list(murfey_url)
        pprint(ongoing_visits)
        ongoing_visits = [v.name for v in ongoing_visits]

    _enable_webbrowser_in_cygwin()

    log.setLevel(logging.DEBUG)
    log_queue = Queue()
    input_queue = Queue()

    # rich_handler = DirectableRichHandler(log_queue, enable_link_path=False)
    rich_handler = DirectableRichHandler(enable_link_path=False)
    rich_handler.setLevel(logging.DEBUG if args.debug else logging.INFO)

    client_id = requests.get(f"{murfey_url.geturl()}/new_client_id/").json()
    ws = murfey.client.websocket.WSApp(
        server=args.server,
        id=client_id["new_id"],
    )

    logging.getLogger().addHandler(rich_handler)
    handler = CustomHandler(ws.send)
    logging.getLogger().addHandler(handler)
    logging.getLogger("murfey").setLevel(logging.INFO)
    logging.getLogger("websocket").setLevel(logging.WARNING)

    log.info("Starting Websocket connection")

    status_bar = StatusBar()

    machine_data = requests.get(f"{murfey_url.geturl()}/machine/").json()
    gain_ref: Path | None = None

    instance_environment = MurfeyInstanceEnvironment(
        url=murfey_url,
        client_id=ws.id,
        software_versions=machine_data.get("software_versions", {}),
        # sources=[Path(args.source)],
        # watchers=source_watchers,
        default_destination=args.destination
        or f"{machine_data.get('rsync_module') or 'data'}/{datetime.now().year}",
        demo=args.demo,
        processing_only_mode=server_routing_prefix_found,
    )

    ws.environment = instance_environment

    rich_handler.redirect = True
    app = MurfeyTUI(
        environment=instance_environment,
        visits=ongoing_visits,
        queues={"input": input_queue, "logs": log_queue},
        status_bar=status_bar,
        dummy_dc=args.fake_dc,
        do_transfer=not args.no_transfer,
        gain_ref=gain_ref,
        redirected_logger=rich_handler,
        force_mdoc_metadata=not args.ignore_mdoc_metadata,
        strict=not args.relax,
        processing_enabled=machine_data.get("processing_enabled", True),
        skip_existing_processing=args.skip_existing_processing,
    )
    app.run()
    rich_handler.redirect = False


def main_loop(
    source_watchers: List[murfey.client.watchdir.DirWatcher],
    appearance_time: float,
    transfer_all: bool,
):
    log.info(
        f"Murfey {murfey.__version__} on Python {'.'.join(map(str, sys.version_info[0:3]))} entering main loop"
    )
    if appearance_time > 0:
        modification_time: float | None = time.time() - appearance_time * 3600
    else:
        modification_time = None
    while True:
        for sw in source_watchers:
            sw.scan(modification_time=modification_time, transfer_all=transfer_all)
        time.sleep(15)


def read_config() -> configparser.ConfigParser:
    config = configparser.ConfigParser()
    try:
        mcch = os.environ.get("MURFEY_CLIENT_CONFIG_HOME")
        murfey_client_config_home = Path(mcch) if mcch else Path.home()
        with open(murfey_client_config_home / ".murfey") as configfile:
            config.read_file(configfile)
    except FileNotFoundError:
        pass
    if "Murfey" not in config:
        config["Murfey"] = {}
    return config


def write_config(config: configparser.ConfigParser):
    mcch = os.environ.get("MURFEY_CLIENT_CONFIG_HOME")
    murfey_client_config_home = Path(mcch) if mcch else Path.home()
    with open(murfey_client_config_home / ".murfey", "w") as configfile:
        config.write(configfile)



from __future__ import annotations

from datetime import datetime
from pathlib import Path

from murfey.util.mdoc import get_block, get_global_data, get_num_blocks


def test_mdoc_file_parse_global_data():
    with open(Path(__file__).parent / "test_1.mdoc", "r") as md:
        data = get_global_data(md)
    assert len(data.keys()) == 6
    assert data["Voltage"] == "300"
    assert data["ImageSize"] == ("4096", "4096")
    assert data["PixelSpacing"] == "1.94"


def test_mdoc_file_parse_block():
    with open(Path(__file__).parent / "test_1.mdoc", "r") as md:
        data = get_block(md)
    assert len(data.keys()) == 32
    assert data["PixelSpacing"] == "1.94"
    assert data["TiltAngle"] == "-0.00949884"
    assert data["DateTime"] == datetime(2022, 8, 1, 18, 58, 35)


def test_mdoc_file_parse_multiple_blocks():
    with open(Path(__file__).parent / "test_1.mdoc", "r") as md:
        data = get_block(md)
        assert len(data.keys()) == 32
        assert data["PixelSpacing"] == "1.94"
        assert data["TiltAngle"] == "-0.00949884"
        assert data["DateTime"] == datetime(2022, 8, 1, 18, 58, 35)
        data = get_block(md)
        assert len(data.keys()) == 32
        assert data["PixelSpacing"] == "1.94"
        assert data["TiltAngle"] == "2.98863"
        assert data["DateTime"] == datetime(2022, 8, 1, 18, 59, 43)


def test_mdoc_file_get_number_of_blocks():
    with open(Path(__file__).parent / "test_1.mdoc", "r") as md:
        block_count = get_num_blocks(md)
    assert block_count == 11



from __future__ import annotations

from fastapi.testclient import TestClient

from murfey.server.main import app

client = TestClient(app)


def test_read_main():
    response = client.get("/")
    assert response.status_code == 200
    assert "<title>" in response.text.lower()


def test_get_visits():
    response = client.get("/visits")
    assert response.status_code == 200
    # assert response.json()[0]["Start date"] == "2020-09-09T14:00:00"


def test_pypi_proxy():
    response = client.get("/pypi/fastapi")
    assert response.status_code == 200
    assert "<a href" in response.text.lower()
    assert ".tar.gz" in response.text.lower()


def test_file_not_found_response():
    response = client.get("/hstnnsv")
    assert response.status_code == 404


def test_openapi_json_is_valid():
    response = client.get("/openapi.json")
    assert response.status_code == 200
    assert response.json()



from __future__ import annotations

from murfey.client.analyser import Analyser


def test_analyser_setup_and_stopping(tmp_path):
    analyser = Analyser(tmp_path)
    assert analyser.queue.empty()
    analyser.start()
    assert analyser.thread.is_alive()
    analyser.stop()
    assert analyser._halt_thread
    assert not analyser.thread.is_alive()


def test_analyser_tomo_determination(tmp_path):
    tomo_file = tmp_path / "Position_1_[30.0].tiff"
    analyser = Analyser(tmp_path)
    analyser.start()
    analyser.queue.put(tomo_file)
    analyser.stop()
    assert analyser._context._acquisition_software == "tomo"


def test_analyser_epu_determination(tmp_path):
    tomo_file = tmp_path / "FoilHole_12345_Data_6789.tiff"
    analyser = Analyser(tmp_path)
    analyser.start()
    analyser.queue.put(tomo_file)
    analyser.stop()
    assert analyser._context._acquisition_software == "epu"
